import json
from typing import Any

import pytest

from tanimachi.schemas import Har


@pytest.fixture
def example():
    with open("tests/fixtures/har/example.com.har") as f:
        return json.load(f)


def test_har(example: Any):
    assert Har.model_validate(example)



from .utils import (  # noqa: F401
    load_categories,
    load_fingerprints,
    load_groups,
    load_har,
)
from .wappalyzer import (  # noqa: F401
    Wappalyzer,
    analyze_css,
    analyze_dom,
    analyze_headers,
    analyze_html,
    analyze_meta,
    analyze_scripts,
    analyze_url,
)



import json
from typing import Any

import pytest

from tanimachi.schemas import Fingerprints


@pytest.fixture
def fingerprints():
    with open("tests/fixtures/wappalyzer/a.json") as f:
        return json.loads(f.read())


def test_fingerprints(fingerprints: Any):
    assert Fingerprints.model_validate(fingerprints)



import itertools
from collections.abc import Callable
from functools import cached_property
from pathlib import Path
from typing import Any, cast

from pydantic import BaseModel
from returns.pipeline import pipe
from selectolax.parser import HTMLParser

from . import schemas
from .utils import load_categories, load_fingerprints, load_groups


def is_javascript(entry: schemas.Entry) -> bool:
    return entry.request.method == "GET" and entry.response.content.mime_type in [
        "application/javascript",
        "text/javascript",
    ]


def is_stylesheet(entry: schemas.Entry) -> bool:
    return entry.request.method == "GET" and entry.response.content.mime_type in [
        "text/css",
    ]


class HarWrapper(schemas.Har):
    @cached_property
    def stylesheet_entries(self):
        return [entry for entry in self.log.entries if is_stylesheet(entry)]

    @cached_property
    def stylesheets(self) -> list[str]:
        stylesheets = [entry.response.content.text for entry in self.stylesheet_entries]
        return [stylesheet for stylesheet in stylesheets if stylesheet]

    @cached_property
    def javascript_entries(self):
        return [entry for entry in self.log.entries if is_javascript(entry)]

    @cached_property
    def scripts(self) -> list[str]:
        scripts = [entry.response.content.text for entry in self.javascript_entries]
        return [script for script in scripts if script]

    @cached_property
    def script_src(self):
        return [entry.request.url for entry in self.javascript_entries]

    @cached_property
    def html_entry(self):
        for entry in self.log.entries:
            if entry.response.content.mime_type == "text/html":
                return entry

        raise ValueError("No HTML entry found")

    @cached_property
    def url(self):
        return str(self.html_entry.request.url)

    @cached_property
    def html(self) -> str:
        if not self.html_entry.response.content.text:
            raise ValueError("No HTML content found")

        return self.html_entry.response.content.text

    @cached_property
    def headers(self):
        return {
            header.name: header.value for header in self.html_entry.response.headers
        }

    @cached_property
    def cookies(self):
        return {
            header.name: header.value for header in self.html_entry.response.cookies
        }

    @cached_property
    def dom(self):
        return HTMLParser(self.html)

    @cached_property
    def meta(self):
        memo: dict[str, str] = {}

        for meta in self.dom.css("meta"):
            name = meta.attributes.get("name")
            content = meta.attributes.get("content")
            if name and content:
                memo[name.lower()] = content

        return memo


class Detection(BaseModel):
    url: str
    fingerprint: schemas.Fingerprint
    app_type: str
    pattern: schemas.Pattern
    value: str
    key: str = ""
    categories: list[schemas.Category] | None = None
    groups: list[schemas.Group] | None = None


def analyze_url(har: HarWrapper, fingerprint: schemas.Fingerprint) -> list[Detection]:
    detections: list[Detection] = []

    for pattern in fingerprint.url:
        if pattern.regex.search(har.url):
            detections.append(
                Detection(
                    url=har.url,
                    fingerprint=fingerprint,
                    app_type="url",
                    pattern=pattern,
                    value=har.url,
                )
            )

    return detections


def analyze_headers(
    har: HarWrapper, fingerprint: schemas.Fingerprint
) -> list[Detection]:
    detections: list[Detection] = []

    for name, patterns in fingerprint.headers.items():
        if name in har.headers:
            content = har.headers[name]
            for pattern in patterns:
                if pattern.regex.search(content):
                    detections.append(
                        Detection(
                            url=har.url,
                            fingerprint=fingerprint,
                            app_type="headers",
                            pattern=pattern,
                            value=content,
                            key=name,
                        )
                    )

    return detections


def analyze_scripts(
    har: HarWrapper, fingerprint: schemas.Fingerprint
) -> list[Detection]:
    detections: list[Detection] = []

    for pattern in fingerprint.scripts:
        for script in har.scripts:
            if pattern.regex.search(script):
                detections.append(
                    Detection(
                        url=har.url,
                        fingerprint=fingerprint,
                        app_type="scripts",
                        pattern=pattern,
                        value=script,
                    )
                )

    return detections


def analyze_css(har: HarWrapper, fingerprint: schemas.Fingerprint) -> list[Detection]:
    detections: list[Detection] = []

    for pattern in fingerprint.css:
        for stylesheet in har.stylesheets:
            if pattern.regex.search(stylesheet):
                detections.append(
                    Detection(
                        url=har.url,
                        fingerprint=fingerprint,
                        app_type="css",
                        pattern=pattern,
                        value=stylesheet,
                    )
                )

    return detections


def analyze_meta(har: HarWrapper, fingerprint: schemas.Fingerprint) -> list[Detection]:
    detections: list[Detection] = []

    for name, patterns in fingerprint.meta.items():
        if name in har.meta:
            content = har.meta[name]
            for pattern in patterns:
                if pattern.regex.search(content):
                    detections.append(
                        Detection(
                            url=har.url,
                            fingerprint=fingerprint,
                            app_type="meta",
                            pattern=pattern,
                            value=content,
                            key=name,
                        )
                    )

    return detections


def analyze_html(har: HarWrapper, fingerprint: schemas.Fingerprint) -> list[Detection]:
    detections: list[Detection] = []

    for pattern in fingerprint.html:
        if pattern.regex.search(har.html):
            detections.append(
                Detection(
                    url=har.url,
                    fingerprint=fingerprint,
                    app_type="html",
                    pattern=pattern,
                    value=har.html,
                )
            )

    return detections


def analyze_dom(har: HarWrapper, fingerprint: schemas.Fingerprint) -> list[Detection]:  # noqa: C901
    detections: list[Detection] = []

    for selector in fingerprint.dom:
        for item in har.dom.css(selector.selector):
            if selector.exists:
                detections.append(
                    Detection(
                        url=har.url,
                        fingerprint=fingerprint,
                        app_type="dom",
                        pattern=schemas.Pattern(string=selector.selector),
                        value="",
                    )
                )

            if selector.text:
                for pattern in selector.text:
                    html = item.html
                    if html and pattern.regex.search(html):
                        detections.append(
                            Detection(
                                url=har.url,
                                fingerprint=fingerprint,
                                app_type="dom",
                                pattern=pattern,
                                value=html,
                            )
                        )

            if selector.attributes:
                for name, patterns in list(selector.attributes.items()):
                    content = item.attributes.get(name)
                    if not content:
                        continue

                    content = str(content)
                    for pattern in patterns:
                        if pattern.regex.search(content):
                            detections.append(
                                Detection(
                                    url=har.url,
                                    fingerprint=fingerprint,
                                    app_type="dom",
                                    pattern=pattern,
                                    value=content,
                                    key=name,
                                )
                            )

    return detections


def filter_by_requires(detections: list[Detection]) -> list[Detection]:
    filtered: list[Detection] = []

    detection_ids = {detection.fingerprint.id for detection in detections}
    for detection in detections:
        if not detection.fingerprint.requires:
            filtered.append(detection)
            continue

        if any(require in detection_ids for require in detection.fingerprint.requires):
            filtered.append(detection)

    return filtered


def filter_by_requires_category(detections: list[Detection]) -> list[Detection]:
    filtered: list[Detection] = []

    categories = set(
        itertools.chain.from_iterable(
            detection.fingerprint.cats for detection in detections
        )
    )
    for detection in detections:
        if not detection.fingerprint.requires_category:
            filtered.append(detection)
            continue

        if any(
            require in categories for require in detection.fingerprint.requires_category
        ):
            filtered.append(detection)

    return filtered


Analyze = Callable[[HarWrapper, schemas.Fingerprint], list[Detection]]


class Wappalyzer:
    def __init__(
        self,
        fingerprints: schemas.Fingerprints | None = None,
        categories: schemas.Categories | None = None,
        groups: schemas.Groups | None = None,
    ):
        self.fingerprints = fingerprints or schemas.Fingerprints(root={})
        self.categories = categories or schemas.Categories(root={})
        self.groups = groups or schemas.Groups(root={})

    def load_fingerprints(self, path: str) -> None:
        """Load fingerprints.

        Args:
            path (str): Glob patterns for the fingerprint files.

        Returns:
            None
        """
        self.fingerprints = load_fingerprints(path)

    def load_groups(self, path: str | Path):
        """Load groups.

        Args:
            path (str | Path): Path to the groups.

        Returns:
            None
        """
        self.groups = load_groups(path)

    def load_categories(self, path: str | Path):
        """Load categories.

        Args:
            path (str | Path): Path to the categories.

        Returns:
            None
        """
        self.categories = load_categories(path)

    def _set_categories(self, detections: list[Detection]):
        if not self.categories:
            return

        for detection in detections:
            memo: dict[str, schemas.Category] = {}

            for category_id in detection.fingerprint.cats:
                category = self.categories.root.get(str(category_id))
                if category:
                    memo[category.id] = category

            if any(memo):
                detection.categories = list(memo.values())

    def _set_groups(self, detections: list[Detection]):
        if not self.groups:
            return

        for detection in detections:
            memo: dict[str, schemas.Group] = {}

            if not detection.categories:
                continue

            for category in detection.categories:
                for group_id in category.groups:
                    group = self.groups.root.get(str(group_id))
                    if group:
                        memo[group.id] = group

            if any(memo):
                detection.groups = list(memo.values())

    def analyze(self, har: schemas.Har | Any, *, analyzes: list[Analyze] | None = None):
        """Analyze HAR.

        Args:
            har (schemas.Har | Any): HAR.

        Returns:
            dict[str, dict[str, str]]: Technologies.
        """
        har = har if isinstance(har, schemas.Har) else schemas.Har.model_validate(har)
        wrapper = HarWrapper(log=har.log)

        analyzes = analyzes or [
            analyze_url,
            analyze_headers,
            analyze_scripts,
            analyze_css,
            analyze_meta,
            analyze_html,
            analyze_dom,
        ]
        detections = itertools.chain.from_iterable(
            itertools.chain.from_iterable(
                [
                    [analyze(wrapper, fingerprint) for analyze in analyzes]
                    for fingerprint in self.fingerprints.root.values()
                ]
            )
        )

        filtered = cast(
            list[Detection],
            pipe(list[Detection], filter_by_requires, filter_by_requires_category)(
                detections
            ),
        )

        self._set_categories(filtered)
        self._set_groups(filtered)

        return filtered



import glob
import json
import tempfile
from typing import Any

import pytest
from git import Repo

from tanimachi import Wappalyzer, schemas
from tanimachi.wappalyzer import (
    analyze_css,
    analyze_headers,
    analyze_html,
    analyze_meta,
    analyze_scripts,
    analyze_url,
)


@pytest.fixture
def repo():
    with tempfile.TemporaryDirectory() as dir:
        Repo.clone_from("https://github.com/tunetheweb/wappalyzer/", dir)
        yield dir


@pytest.fixture
def fingerprints(repo: str):
    memo: dict[str, Any] = {}
    for path in glob.glob(f"{repo}/src/technologies/*.json"):
        with open(path) as f:
            memo.update(json.load(f))

    return schemas.Fingerprints.model_validate(memo)


@pytest.fixture
def har():
    with open("tests/fixtures/har/example.com.har") as f:
        return schemas.Har.model_validate_json(f.read())


def test_integration(har: schemas.Har, fingerprints: schemas.Fingerprints):
    wappalyzer = Wappalyzer(fingerprints=fingerprints)
    # analyze_dom is very slow so skip it
    assert (
        len(
            wappalyzer.analyze(
                har,
                analyzes=[
                    analyze_css,
                    analyze_headers,
                    analyze_html,
                    analyze_meta,
                    analyze_scripts,
                    analyze_url,
                ],
            )
        )
        > 0
    )



import glob
import json
from pathlib import Path
from typing import Any

from . import schemas


def load_har(path: str | Path) -> schemas.Har:
    """Load HAR file.

    Args:
        path (str | Path): Path to the HAR file.

    Returns:
        schemas.Har: HAR.
    """
    path = Path(path) if isinstance(path, str) else path
    return schemas.Har.model_validate_json(path.read_text())


def load_fingerprints(pattern: str) -> schemas.Fingerprints:
    """Load fingerprints.

    Args:
        dir (str): Glob patterns for the fingerprint files.

    Returns:
        schemas.Fingerprint: Fingerprints.
    """
    memo: dict[str, Any] = {}
    for path in glob.glob(pattern):
        memo.update(json.loads(Path(path).read_text()))

    return schemas.Fingerprints.model_validate(memo)


def load_categories(path: str | Path) -> schemas.Categories:
    """Load categories.

    Args:
        path (str | Path): Path to the categories file.

    Returns:
        schemas.Categories: Categories
    """
    path = Path(path) if isinstance(path, str) else path
    return schemas.Categories.model_validate_json(path.read_text())


def load_groups(path: str | Path) -> schemas.Groups:
    """Load groups.

    Args:
        path (str | Path): Path to the groups file.

    Returns:
        schemas.Groups: Groups
    """
    path = Path(path) if isinstance(path, str) else path
    return schemas.Groups.model_validate_json(path.read_text())



import re
from collections.abc import Mapping
from typing import Annotated, Any

from loguru import logger
from pydantic import BaseModel, BeforeValidator, Field, RootModel, model_validator

from .api_model import APIModel


class Pattern(BaseModel):
    string: str
    regex: re.Pattern = re.compile("", 0)
    version: str | None = None
    confidence: int = 100


def prepare_pattern(v: str | list[str]) -> list[Pattern]:
    pattern_objects: list[Pattern] = []

    if isinstance(v, list):
        for p in v:
            pattern_objects.extend(prepare_pattern(p))

        return pattern_objects

    attrs: dict[str, Any] = {}
    patterns = v.split("\\;")
    for index, expression in enumerate(patterns):
        if index == 0:
            attrs["string"] = expression
            try:
                attrs["regex"] = re.compile(expression, re.I)  # type: ignore
            except re.error as e:
                # Wappalyzer is a JavaScript application therefore some of the regex wont compile in Python.
                logger.debug(f"Caught '{e}' compiling regex: {patterns}")
                # regex that never matches:
                # http://stackoverflow.com/a/1845097/413622
                attrs["regex"] = re.compile(r"(?!x)x")  # type: ignore
        else:
            attr = expression.split(":")
            if len(attr) > 1:
                key = attr.pop(0)
                # This adds pattern['version'] when specified with "\\;version:\\1"
                attrs[str(key)] = ":".join(attr)

    pattern_objects.append(Pattern(**attrs))

    return pattern_objects


Patterns = Annotated[list[Pattern], BeforeValidator(prepare_pattern)]


def prepare_pattern_dict(v: dict[str, str | list[str]]) -> Mapping[str, list[Pattern]]:
    memo: dict[str, list[Pattern]] = {}
    for k in v:
        memo[k] = prepare_pattern(v[k])

    return memo


def prepare_headers(v: dict[str, str | list[str]]) -> Mapping[str, list[Pattern]]:
    return prepare_pattern_dict({k.lower(): _v for k, _v in v.items()})


Headers = Annotated[Mapping[str, list[Pattern]], BeforeValidator(prepare_headers)]


def prepare_meta(v: str | list[str] | dict[str, str | list[str]]):
    thing = {"generator": v} if not isinstance(v, dict) else v

    return prepare_pattern_dict({k.lower(): v for k, v in thing.items()})


Meta = Annotated[Mapping[str, list[Pattern]], BeforeValidator(prepare_meta)]


class DomSelector(BaseModel):
    selector: str
    exists: bool | None = None
    text: list[Pattern] | None = None
    attributes: Mapping[str, list[Pattern]] | None = None


def prepare_dom(
    thing: str | list[str] | dict[str, dict[str, str | list[str]]],
) -> list[DomSelector]:
    if isinstance(thing, str):
        return [DomSelector(selector=thing, exists=True)]

    if isinstance(thing, list):
        return [DomSelector(selector=_o, exists=True) for _o in thing]

    selectors: list[DomSelector] = []
    if isinstance(thing, dict):
        for cssselect, clause in thing.items():
            # prepare regexes
            _prep_text_patterns = None
            _prep_attr_patterns = None
            _exists = None
            if clause.get("exists") is not None:
                _exists = True
            if clause.get("text"):
                _prep_text_patterns = prepare_pattern(clause["text"])
            if clause.get("attributes"):
                _prep_attr_patterns = {}
                for _key, pattern in clause["attributes"].items():  # type: ignore
                    _prep_attr_patterns[_key] = prepare_pattern(pattern)
            selectors.append(
                DomSelector(
                    selector=cssselect,
                    exists=_exists,
                    text=_prep_text_patterns,
                    attributes=_prep_attr_patterns,
                )
            )
    return selectors


Dom = Annotated[list[DomSelector], BeforeValidator(prepare_dom)]


def strings(v: str | list[str]) -> list[str]:
    if isinstance(v, list):
        return v

    return [v]


Strings = Annotated[list[str], BeforeValidator(strings)]


def numbers(v: int | list[int]) -> list[int]:
    if isinstance(v, list):
        return v

    return [v]


Numbers = Annotated[list[int], BeforeValidator(numbers)]


class Group(BaseModel):
    id: str = "N/A"
    name: str


class Groups(RootModel):
    root: dict[str, Group]

    @model_validator(mode="after")
    def set_id(self):
        for k, v in self.root.items():
            v.id = k

        return self


class Category(BaseModel):
    id: str = "N/A"
    name: str
    groups: list[int] = Field(default_factory=list)
    priority: int = 0


class Categories(RootModel):
    root: dict[str, Category]

    @model_validator(mode="after")
    def set_id(self):
        for k, v in self.root.items():
            v.id = k

        return self


class Fingerprint(APIModel):
    id: str = "N/A"

    website: str
    cats: list[int] = Field(default_factory=list)
    description: str | None = None
    icon: str | None = None
    cpe: str | None = None
    saas: bool | None = None
    oss: bool | None = None
    pricing: Strings = Field(default_factory=list)

    implies: Strings = Field(default_factory=list)
    requires: Strings = Field(default_factory=list)
    requires_category: Numbers = Field(default_factory=list)
    excludes: Strings = Field(default_factory=list)

    dom: Dom = Field(default_factory=list)

    headers: Headers = Field(default_factory=dict)
    meta: Meta = Field(default_factory=dict)

    html: Patterns = Field(default_factory=list)
    text: Patterns = Field(default_factory=list)
    url: Patterns = Field(default_factory=list)
    script_src: Patterns = Field(default_factory=list)
    scripts: Patterns = Field(default_factory=list)
    css: Patterns = Field(default_factory=list)


class Fingerprints(RootModel):
    root: dict[str, Fingerprint]

    @model_validator(mode="after")
    def set_id(self):
        for k, v in self.root.items():
            v.id = k

        return self



import pytest

from tanimachi import schemas
from tanimachi.wappalyzer import HarWrapper, Wappalyzer, analyze_dom


@pytest.fixture
def har():
    with open("tests/fixtures/har/example.com.har") as f:
        return schemas.Har.model_validate_json(f.read())


@pytest.fixture
def fingerprints():
    with open("tests/fixtures/wappalyzer/a.json") as f:
        return schemas.Fingerprints.model_validate_json(f.read())


def test_analyze(har: schemas.Har, fingerprints: schemas.Fingerprints):
    wappalyzer = Wappalyzer(fingerprints=fingerprints)
    assert len(wappalyzer.analyze(har)) > 0


@pytest.mark.parametrize(
    ("fingerprint", "expected"),
    [
        (schemas.Fingerprint(dom={"h1": {"exists": ""}}, website="dummy"), 1),  # type: ignore
        (schemas.Fingerprint(dom={"h2": {"exists": ""}}, website="dummy"), 0),  # type: ignore
        (
            schemas.Fingerprint(
                dom={"h1": {"text": "Example Domain"}},  # type: ignore
                website="dummy",
            ),
            1,
        ),
        (schemas.Fingerprint(dom={"h1": {"text": "foo"}}, website="dummy"), 0),  # type: ignore
    ],
)
def test_analyze_dom(har: schemas.Har, fingerprint: schemas.Fingerprint, expected: int):
    wrapper = HarWrapper(log=har.log)
    assert len(analyze_dom(wrapper, fingerprint=fingerprint)) == expected



from functools import cached_property

from pydantic import AnyUrl, AwareDatetime, BaseModel, Field

from .api_model import APIModel


class Creator(BaseModel):
    name: str
    version: str
    comment: str | None = None


class Browser(BaseModel):
    name: str
    version: str
    comment: str | None = None


class PageTimings(APIModel):
    on_content_load: float | None = None
    on_load: float | None = None
    comment: str | None = None


class Cookie(APIModel):
    name: str
    value: str
    path: str | None = None
    domain: str | None = None
    expires: AwareDatetime | None = None
    http_only: bool | None = None
    secure: bool | None = None
    comment: str | None = None


class Header(BaseModel):
    name: str
    value: str
    comment: str | None = None


class Query(BaseModel):
    name: str
    value: str
    comment: str | None = None


class Params(APIModel):
    name: str
    value: str | None = None
    file_name: str | None = None
    content_type: str | None = None
    comment: str | None = None


class PostData(APIModel):
    mime_type: str
    text: str | None = None
    params: list | Params | None = None
    comment: str | None = None


class Content(APIModel):
    size: int
    compression: int | None = None
    mime_type: str
    text: str | None = None
    encoding: str | None = None
    comment: str | None = None


class BeforeAfterRequest(APIModel):
    expires: str | None = None
    last_access: str
    e_tag: str
    hit_count: int
    comment: str | None = None


class Timings(BaseModel):
    dns: float | None = None
    connect: float | None = None
    blocked: float | None = None
    send: float
    wait: float
    receive: float
    ssl: float | None = None
    comment: str | None = None


class Page(APIModel):
    started_date_time: AwareDatetime
    id: str
    title: str
    page_timings: PageTimings
    comment: str | None = None


class Request(APIModel):
    method: str
    url: AnyUrl
    http_version: str
    cookies: list[Cookie]
    headers: list[Header]
    query_string: list[Query]
    post_data: PostData | None = None
    headers_size: int
    body_size: int
    comment: str | None = None


class Response(APIModel):
    status: int
    status_text: str
    http_version: str
    cookies: list[Cookie]
    headers: list[Header]
    content: Content
    redirect_url: str = Field(..., alias="redirectURL")
    headers_size: int
    body_size: int
    comment: str | None = None


class Cache(APIModel):
    before_request: BeforeAfterRequest | None = None
    after_request: BeforeAfterRequest | None = None
    comment: str | None = None


class Entry(APIModel):
    pageref: str | None = None
    started_date_time: AwareDatetime
    time: float
    request: Request
    response: Response
    cache: Cache
    timings: Timings
    server_ip_address: str | None = Field(None, alias="serverIPAddress")
    connection: str | None = None
    comment: str | None = None

    @cached_property
    def headers(self) -> dict[str, str]:
        return {header.name: header.value for header in self.response.headers}

    @cached_property
    def cookies(self) -> dict[str, str]:
        return {header.name: header.value for header in self.response.cookies}


class Log(BaseModel):
    version: str
    creator: Creator
    browser: Browser | None = None
    pages: list[Page] | None = None
    entries: list[Entry]
    comment: str | None = None


class Har(BaseModel):
    log: Log
